@@ -1,7 +1,9 @@
 import collections
+import itertools
 import tempfile
 import shutil
 import subprocess
+import logging
 import os
 import sys
 import re
@@ -12,6 +14,9 @@ from flask import Flask, request, redirect, send_file, send_from_directory, rend
 import os.path
 import requests
 
+logging.basicConfig(level=logging.DEBUG)
+log = logging.getLogger(__name__ )
+log.debug("Entering web uploader")
 
 if not os.path.isfile('bh20sequploader/mainx.py'):
     print("WARNING: run FLASK from the root of the source repository!", file=sys.stderr)
@@ -47,10 +52,10 @@ def name_to_label(field_name):
     """
     Turn a filed name like "host_health_status" from the metadata schema into a human-readable label.
     """
-    
+
     # May end in a number, which should be set off by a space
     set_off_number = re.sub('([0-9]+)$', r' \1', field_name)
-    
+
     return string.capwords(set_off_number.replace('_', ' '))
 
 def is_iri(string):
@@ -71,13 +76,13 @@ def generate_form(schema, options):
     Each dict either has a 'heading' (in which case we put a heading for a
     form section in the template) or an 'id', 'label', 'type', and 'required'
     (in which case we make a form field in the template).
-    
+
     Non-heading dicts with type 'select' will have an 'options' field, with a
     list of (name, value) tuples, and represent a form dropdown element.
-    
+
     Non-heading dicts with type 'number' may have a 'step', which, if <1 or
     'any', allows the number to be a float.
-    
+
     Non-heading dicts may have a human-readable 'docstring' field describing
     them.
 
@@ -124,7 +129,7 @@ def generate_form(schema, options):
             docstring = None
             if not isinstance(field_type, str):
                 # If the type isn't a string
-                
+
                 # It may have documentation
                 docstring = field_type.get('doc', None)
 
@@ -153,12 +158,20 @@ def generate_form(schema, options):
 
             # Decide if the field is optional (type ends in ?)
             optional = False
-            if len(field_type) > 0 and field_type[-1] == '?':
+            if field_type.endswith('?'):
                 # It's optional
                 optional = True
                 # Drop the ?
                 field_type = field_type[:-1]
 
+            # Decide if the field is a list (type ends in [])
+            is_list = False
+            if field_type.endswith('[]'):
+                # It's a list
+                is_list = True
+                # Reduce to the normal type
+                field_type = field_type[:-2]
+
             if field_type in by_name:
                 # This is a subrecord. We need to recurse
                 for item in walk_fields(field_type, parent_keys + [field_name], subtree_optional or optional):
@@ -169,6 +182,7 @@ def generate_form(schema, options):
                 record['id'] = '.'.join(parent_keys + [field_name])
                 record['label'] = name_to_label(field_name)
                 record['required'] = not optional and not subtree_optional
+                record['list'] = is_list
                 if ref_iri:
                     record['ref_iri'] = ref_iri
                 if docstring:
@@ -193,18 +207,10 @@ def generate_form(schema, options):
                         record['type'] = 'text'
                 elif field_type == 'int':
                     record['type'] = 'number'
-                elif field_type == 'float':
+                elif field_type == 'float' or field_type == 'double':
                     record['type'] = 'number'
                     # Choose a reasonable precision for the control
                     record['step'] = '0.0001'
-
-                ### This is to fix the homepage for the moment ## needs more love though
-                # implementation of the [] stuff instead of just text fields
-                ## ToDo - implement lists
-                elif field_type == 'string[]':
-                    record['type'] = 'text'
-                elif field_type == 'float[]':
-                    record['type'] = 'text'
                 else:
                     raise NotImplementedError('Unimplemented field type {} in {} in metadata schema'.format(field_type, type_name))
                 yield record
@@ -223,7 +229,7 @@ def send_form():
     Send the file upload form/front page.
     """
 
-    return render_template('form.html', fields=FORM_ITEMS)
+    return render_template('form.html', fields=FORM_ITEMS, menu='HOME')
 
 class FileTooBigError(RuntimeError):
     """
@@ -248,18 +254,27 @@ def copy_with_limit(in_file, out_file, limit=1024*1024):
         buf = in_file.read(buf_size)
         bytes_used += len(buf)
 
-def parse_input(input_string, html_type):
+def parse_input(input_string, html_type, number_step=None):
     """
     Parse an input from the given HTML input type into a useful Python type.
+    Also needs the step we sent to distinguish int fields and float/double fields.
 
     Raise ValueError if something does not parse.
     Raise NotImplementedError if we forgot to implement a type.
     """
 
-    if html_type == 'text':
+    if html_type == 'text' or html_type == 'select':
         return input_string
     elif html_type == 'number':
-        return int(input_string)
+        # May be an int or a float.
+        if number_step is None:
+            # TODO: Assumes we only use the step for floats
+            return int(input_string)
+        else:
+            return float(input_string)
+    elif html_type == 'date':
+        # Don't do our own date validation; pass it on as a string
+        return input_string
     else:
         raise NotImplementedError('Unimplemented input type: {}'.format(html_type))
 
@@ -300,30 +315,83 @@ def receive_files():
             # Build a metadata dict
             metadata = {}
 
+            # When we have metadata for an item, use this to set it.
+            # If it is an item in a list, set is_list=True
+            def set_metadata(item_id, value, is_list=False):
+                # We have this thing. Make a place in the dict tree for it.
+                parts = item_id.split('.')
+                key = parts[-1]
+                # Remove leading 'metadata'
+                path = parts[1:-1]
+                dest_dict = metadata
+                for parent in path:
+                    if parent not in dest_dict:
+                        dest_dict[parent] = {}
+                    dest_dict = dest_dict[parent]
+
+                if not is_list:
+                    dest_dict[key] = value
+                else:
+                    if key not in dest_dict:
+                        dest_dict[key] = []
+                    dest_dict[key].append(value)
+
             for item in FORM_ITEMS:
                 # Pull all the field values we wanted from the form
                 if 'heading' in item:
                     continue
 
-                if item['id'] in request.form and len(request.form[item['id']]) > 0:
-                    # We have this thing. Make a place in the dict tree for it.
-                    parts = item['id'].split('.')
-                    key = parts[-1]
-                    # Remove leading 'metadata'
-                    path = parts[1:-1]
-                    dest_dict = metadata
-                    for parent in path:
-                        if parent not in dest_dict:
-                            dest_dict[parent] = {}
-                        dest_dict = dest_dict[parent]
+                if item['list']:
+                    # This is a list, serialized into form fields
 
+                    # We count how many values we got
+                    value_count = 0
+
+                    for index in itertools.count():
+                        # Get [0] through [n], until something isn't there.
+                        entry_id = '{}[{}]'.format(item['id'], index)
+
+                        if index == 1000:
+                            # Don't let them provide too much stuff.
+                            return (render_template('error.html',
+                                    error_message="You provided an extremely large number of values for the metadata item {}".format(item['id'])), 403)
+
+                        if entry_id in request.form:
+                            if len(request.form[entry_id]) > 0:
+                                # Put an entry in the list
+                                try:
+                                    # Parse the item
+                                    parsed = parse_input(request.form[entry_id], item['type'], item.get('step', None))
+                                except ValueError:
+                                    # We don't like that input
+                                    return (render_template('error.html',
+                                            error_message="You provided an unacceptable value for the metadata item {}".format(entry_id)), 403)
+                                # Save it
+                                set_metadata(item['id'], parsed, is_list=True)
+                                value_count += 1
+                            else:
+                                # Empty items are silently skipped.
+                                pass
+                        else:
+                            # We have run out of form fields for this list.
+                            break
+
+                        if item['required'] and value_count == 0:
+                            # They forgot a required item. Maybe all entries were empty.
+                            return (render_template('error.html',
+                                    error_message="You omitted any values for the required metadata item {}".format(item['id'])), 403)
+
+                elif item['id'] in request.form and len(request.form[item['id']]) > 0:
+                    # Not a list, but a single item which is present.
                     try:
-                        # Now finally add the item
-                        dest_dict[key] = parse_input(request.form[item['id']], item['type'])
+                        # Parse the item
+                        parsed = parse_input(request.form[item['id']], item['type'], item.get('step', None))
                     except ValueError:
                         # We don't like that input
                         return (render_template('error.html',
                             error_message="You provided an unacceptable value for the metadata item {}".format(item['id'])), 403)
+                    # Save it
+                    set_metadata(item['id'], parsed)
                 elif item['required']:
                     return (render_template('error.html',
                             error_message="You omitted the required metadata item {}".format(item['id'])), 403)
@@ -353,6 +421,50 @@ def receive_files():
     finally:
         shutil.rmtree(dest_dir)
 
+def get_html_body(fn):
+    buf = ""
+    in_body = False
+    begin_body = re.compile(r"<body>",re.IGNORECASE)
+    end_body = re.compile(r"(</body>|.*=\"postamble\")",re.IGNORECASE)
+    with open(fn) as f:
+        for line in f:
+            if end_body.match(line):
+                break
+            if in_body:
+                buf += line
+            elif begin_body.match(line):
+                in_body = True
+    return buf
+
+@app.route('/download')
+def download_page():
+    buf = get_html_body('doc/web/download.html')
+    return render_template('about.html',menu='DOWNLOAD',embed=buf)
+
+@app.route('/demo')
+def demo_page():
+    return render_template('demo.html',menu='DEMO')
+
+@app.route('/blog',methods=['GET'])
+def blog_page():
+    blog_content = request.args.get('id') # e.g. using-covid-19-pubseq-part3
+    buf = None;
+    if blog_content:
+        buf = get_html_body('doc/blog/'+blog_content+'.html')
+    return render_template('blog.html',menu='BLOG',embed=buf,blog=blog_content)
+
+
+@app.route('/about')
+def about_page():
+    buf = get_html_body('doc/web/about.html')
+    return render_template('about.html',menu='ABOUT',embed=buf)
+
+##
+@app.route('/map')
+def map_page():
+    return render_template('map.html',menu='DEMO')
+
+
 
 ## Dynamic API functions starting here
 ## This is quick and dirty for now, just to get something out and demonstrate the queries
@@ -360,6 +472,23 @@ def receive_files():
 ## but most likley you don't want to touch the queries, Cheers.
 baseURL='http://sparql.genenetwork.org/sparql/'
 
+@app.route('/api/getCount', methods=['GET'])
+def getCount():
+    query="""
+PREFIX pubseq: <http://biohackathon.org/bh20-seq-schema#MainSchema/>
+select (COUNT(distinct ?dataset) as ?num)
+{
+   ?dataset pubseq:submitter ?id .
+   ?id ?p ?submitter
+}
+"""
+    payload = {'query': query, 'format': 'json'}
+    r = requests.get(baseURL, params=payload)
+    result = r.json()['results']['bindings']
+    # [{'num': {'type': 'typed-literal', 'datatype': 'http://www.w3.org/2001/XMLSchema#integer', 'value': '1352'}}]
+    # print(result, file=sys.stderr)
+    return jsonify({'sequences': int(result[0]["num"]["value"])})
+
 @app.route('/api/getAllaccessions', methods=['GET'])
 def getAllaccessions():
     query="""SELECT DISTINCT ?fasta ?value WHERE {?fasta ?x[ <http://edamontology.org/data_2091> ?value ]}"""
@@ -371,50 +500,61 @@ def getAllaccessions():
 
 
 # parameter must be encoded e.g. http://arvados.org/keep:6e6276698ed8b0e6cd21f523e4f91179+123/sequence.fasta must become
-# http%3A%2F%2Farvados.org%2Fkeep%3A6e6276698ed8b0e6cd21f523e4f91179%2B123%2Fsequence.fasta
+# http%3A%2F%2Fcollections.lugli.arvadosapi.com%2Fc%3D00a6af865453564f6a59b3d2c81cc7c1%2B123%2Fsequence.fasta
 @app.route('/api/getDetailsForSeq', methods=['GET'])
 def getDetailsForSeq():
     seq_id = request.args.get('seq')
-    query="""SELECT DISTINCT ?key ?value WHERE {<placeholder> ?x [?key ?value]}"""
+    query="""SELECT DISTINCT ?key ?key_label ?value WHERE {
+    <placeholder> ?x [?key ?value] .
+    OPTIONAL {?key <http://www.w3.org/2000/01/rdf-schema#label> ?key_tmp_label } .
+    BIND(IF(BOUND(?key_tmp_label),?key_tmp_label, ?key) as ?key_label)}"""
     query=query.replace("placeholder", seq_id)
     payload = {'query': query, 'format': 'json'}
     r = requests.get(baseURL, params=payload)
     result = r.json()['results']['bindings']
-    return jsonify([{'uri': x['key']['value'],
+    return jsonify([{'uri': x['key']['value'], 'key_label': x['key_label']['value'],
                      'value': x['value']['value']} for x in result])
 
+# Endpoint should provide all necessary information to draw a map (!)
+@app.route('/api/getCountByGPS', methods=['GET'])
+def getCountByGPS():
+    query="""SELECT DISTINCT ?location ?location_label ?GPS (count(?fasta) as ?fastaCount) WHERE {
+    ?fasta ?x[ <http://purl.obolibrary.org/obo/GAZ_00000448> ?location] .
+    ?location <http://www.wikidata.org/prop/direct/P625> ?GPS .
+    OPTIONAL { ?location rdfs:label ?key_tmp_label }
+    BIND(IF(BOUND(?key_tmp_label),?key_tmp_label, ?location) as ?location_label)
+    }
+    GROUP BY ?location ?location_label ?GPS
+    """
+    payload = {'query': query, 'format': 'json'}
+    r = requests.get(baseURL, params=payload)
+    result = r.json()['results']['bindings']
+    return jsonify([{'count': x['fastaCount']['value'],
+                     'Location': x['location']['value'],
+                     'LocationLabel': x['location_label']['value'],
+                     'GPS' :x['GPS']['value'][6:-1]} for x in result])
+
 
 @app.route('/api/getSEQCountbytech', methods=['GET'])
 def getSEQCountbytech():
-    query="""SELECT ?tech ?tech_label (count(?fasta) as ?fastaCount) WHERE 
-    {?fasta ?x [<http://purl.obolibrary.org/obo/OBI_0600047>  ?tech] 
-    BIND (concat(?tech,"_label") as ?tech_label)}
+    query="""SELECT ?tech ?tech_label (count(?fasta) as ?fastaCount) WHERE
+    {?fasta ?x [<http://purl.obolibrary.org/obo/OBI_0600047>  ?tech] .
+     OPTIONAL {?tech <http://www.w3.org/2000/01/rdf-schema#label> ?tech_tmp_label } .
+     BIND(IF(BOUND(?tech_tmp_label), ?tech_tmp_label,?tech) as ?tech_label)}
     GROUP BY ?tech ?tech_label ORDER BY DESC (?fastaCount)
     """
     payload = {'query': query, 'format': 'json'}
     r = requests.get(baseURL, params=payload)
     result = r.json()['results']['bindings']
-    return jsonify([{'Fasta Count': x['fastaCount']['value'],
-                     'tech': x['tech']['value'],
-                     'Label': x['tech_label']['value']} for x in result])
-
-## Is this one really necessary or should we just use getSEQCountbytech instead?
-@app.route('/api/getAvailableTech', methods=['GET'])
-def getAvailableTech():
-    query="""SELECT distinct ?tech ?tech_label WHERE 
-    {?fasta ?x [<http://purl.obolibrary.org/obo/OBI_0600047> ?tech] 
-     BIND (concat(?tech,"_label") as ?tech_label)
-    } """
-    payload = {'query': query, 'format': 'json'}
-    r = requests.get(baseURL, params=payload)
-    result = r.json()['results']['bindings']
-    return str(result)
+    return jsonify([{'count': x['fastaCount']['value'],
+                     'key': x['tech']['value'],
+                     'label': x['tech_label']['value']} for x in result])
 
 ## List all Sequences/submissions by a given tech, as example e.g. http://purl.obolibrary.org/obo/OBI_0000759
 ## Has to be encoded again so should be --> http%3A%2F%2Fpurl.obolibrary.org%2Fobo%2FOBI_0000759
 @app.route('/api/getSEQbytech', methods=['GET'])
 def getSEQbytech():
-    query="""SELECT ?fasta WHERE 
+    query="""SELECT ?fasta WHERE
     {?fasta ?x [<http://purl.obolibrary.org/obo/OBI_0600047>  <placeholder>] }
     """
     tech = request.args.get('tech')
@@ -441,32 +581,75 @@ def getSEQbyLocation():
 @app.route('/api/getSEQCountbyLocation', methods=['GET'])
 def getSEQCountbyLocation():
     query="""SELECT ?geoLocation ?geoLocation_label (count(?fasta) as ?fastaCount)  WHERE
-    {?fasta ?x [<http://purl.obolibrary.org/obo/GAZ_00000448> ?geoLocation]
-    BIND (concat(?geoLocation,"_label") as ?geoLocation_label)}
+    {?fasta ?x [<http://purl.obolibrary.org/obo/GAZ_00000448> ?geoLocation] .
+    Optional {?geoLocation <http://www.w3.org/2000/01/rdf-schema#label> ?geoLocation_tmp_label}
+    BIND(IF(BOUND(?geoLocation_tmp_label), ?geoLocation_tmp_label, ?geoLocation) as ?geoLocation_label)}
     GROUP BY ?geoLocation ?geoLocation_label ORDER BY DESC (?fastaCount)
     """
     payload = {'query': query, 'format': 'json'}
     r = requests.get(baseURL, params=payload)
     result = r.json()['results']['bindings']
-    return jsonify([{'Fasta Count': x['fastaCount']['value'],
-                     'GeoLocation': x['geoLocation']['value'],
-                     'GeoLocation Label': x['geoLocation_label']['value']} for x in result])
+    return jsonify([{'count': x['fastaCount']['value'],
+                     'key': x['geoLocation']['value'],
+                     'label': x['geoLocation_label']['value']} for x in result])
+
+
+@app.route('/api/getSEQCountbyContinent', methods=['GET'])
+def getSEQCountbyContinent():
+    query="""SELECT DISTINCT ?continent ?continent_label (count(?fasta) as ?fastaCount) WHERE {
+    ?fasta ?x[ <http://purl.obolibrary.org/obo/GAZ_00000448> ?location] .
+   ?location <http://www.wikidata.org/prop/direct/P17> ?country .
+    ?country <http://www.wikidata.org/prop/direct/P30> ?continent .
+    OPTIONAL { ?continent rdfs:label ?key_tmp_label }
+    BIND(IF(BOUND(?key_tmp_label),?key_tmp_label, ?location) as ?continent_label)
+    }
+    GROUP BY ?continent ?continent_label
+    """
+    payload = {'query': query, 'format': 'json'}
+    r = requests.get(baseURL, params=payload)
+    result = r.json()['results']['bindings']
+    return jsonify([{'count': x['fastaCount']['value'],
+                     'key': x['continent']['value'],
+                     'label': x['continent_label']['value']} for x in result])
+
+
+
+@app.route('/api/getSEQCountbyCountryContinent', methods=['GET'])
+def getSEQCountbyCountryContinent():
+    query="""SELECT DISTINCT ?location ?location_label (count(?fasta) as ?fastaCount) WHERE {
+    ?fasta ?x[ <http://purl.obolibrary.org/obo/GAZ_00000448> ?location] .
+    ?location <http://www.wikidata.org/prop/direct/P30> <placeholder> .
+    OPTIONAL { ?location rdfs:label ?key_tmp_label }
+    BIND(IF(BOUND(?key_tmp_label),?key_tmp_label, ?location) as ?location_label)
+    }
+    GROUP BY ?location ?location_label
+    """
+    continent = request.args.get('continent')
+    query = query.replace("placeholder", continent)
+    payload = {'query': query, 'format': 'json'}
+    r = requests.get(baseURL, params=payload)
+    result = r.json()['results']['bindings']
+    return jsonify([{'count': x['fastaCount']['value'],
+                     'key': x['location']['value'],
+                     'label': x['location_label']['value']} for x in result])
+
 
 
 @app.route('/api/getSEQCountbySpecimenSource', methods=['GET'])
 def getSEQCountbySpecimenSource():
     query="""SELECT ?specimen_source ?specimen_source_label (count(?fasta) as ?fastaCount)  WHERE
     {?fasta ?x [<http://purl.obolibrary.org/obo/OBI_0001479>  ?specimen_source]
-    BIND (concat(?specimen_source,"_label") as ?specimen_source_label)}
-    GROUP BY ?specimen_source ?specimen_source_label
-    ORDER BY DESC (?fastaCount)
+     Optional { ?specimen_source <http://www.w3.org/2000/01/rdf-schema#label> ?specimen_source_tmp_label}
+     BIND(IF(BOUND(?specimen_source_tmp_label), ?specimen_source_tmp_label ,?specimen_source) as ?specimen_source_label)}
+     GROUP BY ?specimen_source ?specimen_source_label
+     ORDER BY DESC (?fastaCount)
     """
     payload = {'query': query, 'format': 'json'}
     r = requests.get(baseURL, params=payload)
     result = r.json()['results']['bindings']
-    return jsonify([{'Fasta Count': x['fastaCount']['value'],
-                     'Specimen Source': x['specimen_source']['value'],
-                     'Label': x['specimen_source_label']['value']} for x in result])
+    return jsonify([{'count': x['fastaCount']['value'],
+                     'key': x['specimen_source']['value'],
+                     'label': x['specimen_source_label']['value']} for x in result])
 
 # Example specimen http%3A%2F%2Fpurl.obolibrary.org%2Fobo%2FNCIT_C155831
 @app.route('/api/getSEQbySpecimenSource', methods=['GET'])
@@ -498,7 +681,7 @@ def getSEQCountbyHostHealthStatus():
 
 @app.route('/api/getSEQbyLocationAndTech', methods=['GET'])
 def getSEQbyLocationAndTech():
-    query="""SELECT ?fasta WHERE { ?fasta ?x [ 
+    query="""SELECT ?fasta WHERE { ?fasta ?x [
         <http://purl.obolibrary.org/obo/GAZ_00000448> <placeholderLoc>; <http://purl.obolibrary.org/obo/OBI_0600047>  <placeholderTech> ]}"""
     location=request.args.get('location')
     tech=request.args.get('tech')
@@ -515,7 +698,7 @@ def getSEQbyLocationAndTech():
 # Example specimen http%3A%2F%2Fpurl.obolibrary.org%2Fobo%2FNCIT_C155831
 @app.route('/api/getSEQbyLocationAndSpecimenSource', methods=['GET'])
 def getSEQbyLocationAndSpecimenSource():
-    query="""SELECT ?fasta WHERE { ?fasta ?x [ 
+    query="""SELECT ?fasta WHERE { ?fasta ?x [
         <http://purl.obolibrary.org/obo/GAZ_00000448> <placeholderLoc>; <http://purl.obolibrary.org/obo/OBI_0001479>  <placeholderSpecimen> ]}
     """
     location = request.args.get('location')
@@ -526,4 +709,4 @@ def getSEQbyLocationAndSpecimenSource():
     payload = {'query': query, 'format': 'json'}
     r = requests.get(baseURL, params=payload)
     result = r.json()['results']['bindings']
-    return str(result)
\ No newline at end of file
+    return str(result)
